var fs = require('fs');
var path = require('path');
var marked = require('marked');
var tempos = require('tempos/src');
var yaml = require('js-yaml');
var gulp = require('gulp');
var sass = require('gulp-ruby-sass');
var autoprefixer = require('autoprefixer');
var nano = require('gulp-cssnano');
var postcss = require('gulp-postcss');
var uglify = require('uglify-js');
var imagemin = require('gulp-imagemin');
var rename = require('gulp-rename');
var concat = require('gulp-concat');
var base64 = require('gulp-base64');
var rollup = require('rollup');
var json = require('rollup-plugin-json');
var babel = require('rollup-plugin-babel');
var resolve = require('rollup-plugin-node-resolve');
var replace = require('rollup-plugin-replace');
var commonjs = require('rollup-plugin-commonjs');
var browserSync = require('browser-sync');
var extend = require('extend');

var RYAML = /^----([\s\S]+?)----([\s\S]+)/;
var base = __dirname;
var option = {};

//
var cfg = option.config = yamlLoad('./_config.yml').yaml;
cfg.website = cfg.protocol + '://' + cfg.host + ':' + cfg.port + cfg.root;
cfg.pro_name = path.parse(path.resolve('.')).name;
cfg.theme_dir = './theme/' + cfg.theme;
cfg.source_dir = cfg.theme_dir + '/source';

createDir('./post');
createDir('./public');
option.layout = getFiles( cfg.theme_dir + '/template' );
option.post = getFiles('post', '.md', true);

// html
gulp.task('html', function() {
    for( var key in option.layout ) {
        var it = option.layout[key], postHook, cur_opt;
        var paths = path.resolve(base, cfg.generate_dir + '/' + it.filename);

        it.compile = tempos.compile(null, it);
        
        if( postHook = option.post[ it.filename ] ) {
            createDir(paths);
            for( var k in postHook ) {
                cur_opt = extend({}, option, it, postHook[k]);
                writeFile(cur_opt, path.resolve(paths, cur_opt.filename + '.html'));
            }
        }
        else {
            cur_opt = extend({}, option, it);
            writeFile(cur_opt, paths + '.html');
        }
    }
});


// copy
gulp.task('copy', function() {
    gulp.src(cfg.source_dir+'/image/**/*')
        .pipe(imagemin({
            optimizationLevel: 7,
            progressive: true,
            interlaced: true
        }))
        .pipe(gulp.dest(cfg.generate_dir + '/image'));
        
    gulp.src(cfg.source_dir+'/plugin/**/*')
        .pipe(gulp.dest(cfg.generate_dir));
});


// styles
gulp.task('styles', function() {
	return sass(cfg.source_dir + '/styles/index.scss')
		.pipe(concat(cfg.pro_name + '.css'))
		.pipe(postcss([
			autoprefixer({
				browsers: [
					'last 2 versions',
					'last 2 Chrome versions',
					'> 5%',
					'> 5% in US',
					'ie >= 9',
					'Firefox >= 20',
					'Firefox <= 20',
					'iOS 7'
				]
			})
		]))
		.pipe(base64())
		.pipe(nano())
		.pipe(gulp.dest(cfg.generate_dir + '/styles'));
});


// scripts
gulp.task('scripts', function() {
    var rollup_cfg = {
        entry: cfg.source_dir + '/scripts/index.js',
        dest: cfg.generate_dir + '/scripts/'+ cfg.pro_name +'.js',
        format: 'umd',
        moduleName: cfg.pro_name,
        sourceMap: true,
        plugins: [
            json(),
            resolve({
                jsnext: true,
                main: true
            }),
            commonjs({
                ignoreGlobal: true
            }),
            babel({
                exclude: ['node_modules/**'],
                presets: ['es2015-rollup']
            }),
            replace({
                exclude: 'node_modules/**',
                ENV: JSON.stringify(process.env.NODE_ENV || 'development')
            })
        ]
    };

    createDir(cfg.generate_dir+'/scripts/');
    rollup.rollup(rollup_cfg).then( (bundle) => {
        var code = bundle.generate(rollup_cfg).code;
        var res = process.env.NODE_ENV === 'production' ? uglify.minify(code, {
            output: {
                ascii_only: true
            },
            compress: {
                pure_funcs: ['makeMap']
            }
        }).code : code;

        fs.writeFileSync(rollup_cfg.dest, res, 'utf8');
    });
});


// default task
gulp.task('default', function() {
	gulp.start(['copy', 'html', 'styles', 'scripts']);
});


// watch
gulp.task('watch', function() {
    browserSync.init({
        port: cfg.port,
        server: {
            baseDir: cfg.generate_dir,
            directory: true
        },
        reloadDebounce: 0
    });
    
    gulp.watch([
        './theme/'+cfg.theme+'/**/*',
        './post/**/*'
    ], ['html', 'styles', 'scripts'], browserSync.reload);
});


// get directory
function getDir(paths){
    var res;
    paths = path.resolve(base, paths);
    try{
        res = fs.readdirSync(paths);
    } catch(err) {
        throw `Directory "${paths}"  does not exist.`;
    }
    return res;
}


// create dir
function createDir(paths) {
    if( typeof paths !== 'string' ) {
        throw 'createDir: The parameter "paths" must be a string.';
    }
    var pa = paths.split(/\\/g);
    var len = pa.length;
    var index = 0;
    while( ++index <= len ) {
        var curp = pa.slice(0, index).join('\\');
        if( !fs.existsSync(curp) ) {
            fs.mkdirSync(curp);
        }
    }
}


// load yaml file
function yamlLoad(paths, pattern) {
    var res = {}, mat;
    var  file = fs.readFileSync( path.resolve(base, paths) ).toString();
    if( pattern ) {
        if( mat = file.match(pattern) ) {
            res.yaml = yaml.safeLoad(mat[1]);
            res.template = mat[2].trim();
        }
        else {
            res.template = file;
        }
    }
    else {
        res.yaml = yaml.load(file);
    }
    return res;
}


// get files
// return all files while all is true
function getFiles (paths, suffix, all) {
    var dir = getDir(paths);
    var res = {};
    suffix = suffix || '.temp';

    dir.map( (f) => {
        var curPath = path.resolve(base, paths + '/' + f);
        var isDir = fs.statSync(curPath).isDirectory();
        if( !isDir ) {
            var r = f.split(/\./);
            var tmp = {};

            tmp.extname = ( '.' + r.pop() ).trim();
            tmp.filename = r.join('');
            tmp.filepath = curPath;
            
            if( !suffix || suffix === tmp.extname ) {
                var mat = yamlLoad(curPath, RYAML);
                extend(tmp, mat.yaml || {});
                tmp.template = mat.template || '';

                if( '.md' === suffix ) {
                    tmp.template = marked(tmp.template);
                }
                tmp.source = tmp.template;

                res[ tmp.filename ] = tmp;
            }
        }
        else if(all) {
            res[ f ] = getFiles(curPath, suffix, all);
        }
    });

    return res;
}


// generated file
function writeFile(res, paths) {
    fs.writeFileSync(paths, res.compile(res), 'utf8');
}